
#ifndef PFMPREFIX_H
#define PFMPREFIX_H
#include "ptpublic.h"

#define PFM_COMPANY                 Pismo Technic Inc.
#define PFM_COMPANYA               "Pismo Technic Inc."
#define PFM_COMPANYW              L"Pismo Technic Inc."
#define PFM_REVURL                  com.pismotechnic
#define PFM_REVURLA                "com.pismotechnic"
#define PFM_REVURLW               L"com.pismotechnic"
#define PFM_PRODUCT                 Pismo File Mount
#define PFM_PRODUCTA               "Pismo File Mount"
#define PFM_PRODUCTW              L"Pismo File Mount"
#define PFM_PRODID                  PismoFileMount
#define PFM_PRODIDA                "PismoFileMount"
#define PFM_PRODIDW               L"PismoFileMount"
#define PFM_COPYRIGHT               Pismo Technic Inc. Copyright 2006-2015 Joe Lowe
#define PFM_COPYRIGHTA             "Pismo Technic Inc. Copyright 2006-2015 Joe Lowe"
#define PFM_COPYRIGHTW            L"Pismo Technic Inc. Copyright 2006-2015 Joe Lowe"
#define PFM_DATE                    2015.3.25
#define PFM_DATEA                  "2015.3.25"
#define PFM_DATEW                 L"2015.3.25"
#define PFM_BUILDTAG                pfm.1.0.0.174
#define PFM_BUILDTAGA              "pfm.1.0.0.174"
#define PFM_BUILDTAGW             L"pfm.1.0.0.174"
#define PFM_BUILD                   174
#define PFM_BUILDA                 "174"
#define PFM_BUILDW                L"174"
#define PFM_PREFIX                  pfm
#define PFM_PREFIXA                "pfm"
#define PFM_PREFIXW               L"pfm"
#define PFM_APIID                   pfmapi
#define PFM_APIIDA                 "pfmapi"
#define PFM_APIIDW                L"pfmapi"
#define PFM_APIBASENAME             pfmapi_174
#define PFM_APIBASENAMEA           "pfmapi_174"
#define PFM_APIBASENAMEW          L"pfmapi_174"
#define PFM_KERNELBASENAME          pfmfs_174
#define PFM_KERNELBASENAMEA        "pfmfs_174"
#define PFM_KERNELBASENAMEW       L"pfmfs_174"
#define PFM_CMDBASENAME             pfm
#define PFM_CMDBASENAMEA           "pfm"
#define PFM_CMDBASENAMEW          L"pfm"

#endif
